// File: CORBA_DomainManager.idl
// CORBA 3.0, Chapter 4

#ifdef _PRE_3_0_COMPILER_
#pragma prefix "omg.org/CORBA"
#endif

    interface DomainManager {
        Policy get_domain_policy (
            in PolicyType           policy_type
        );
    };

    const PolicyType SecConstruction = 11;

    interface ConstructionPolicy: Policy {
        void make_domain_manager(
            in CORBA::InterfaceDef  object_type,
            in boolean              constr_policy
        );
    };

    typedef sequence <DomainManager> DomainManagersList;

